/*-
 * Copyright (c) 2013 Devin Teske <dteske@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <err.h>
#include <stdarg.h>
#define _BSD_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

#include "dialog.h"
#include "dprompt.h"
#include "fdpv.h"
#include "string_m.h"

#define FLABEL_MAX 1024

static char dprompt[PROMPT_MAX + 1] = "";
static char *dprompt_pos = (char *)(0); /* treated numerically */
static int fheight = 0; /* initialized by dprompt_init() */

/* Display characteristics */
int label_size = 28;	/* Maximum width for labels (file names) */
int pbar_size = 17;	/* How big is the mini-progressbar? */
#define SPIN_SIZE 4
static char spin[SPIN_SIZE + 1] = "/-\\|";
static char *spin_cp = spin;
static int pct_lsize, pct_rsize;
static int done_lsize, done_rsize;
#define PENDING "Pending"
#define PENDING_SIZE 7
static char pending[PENDING_SIZE + 1];
static int pend_lsize, pend_rsize;

/* Function prototypes for private functions (see style(9)) */
char spin_char(void);

/*
 * Returns a pointer to the current spin character in the spin string and
 * advances the global position to the next character for the next call.
 */
char
spin_char(void)
{
	char ch;

	if (spin_cp == '\0') spin_cp = spin;
	ch = *spin_cp;

	/* Advance the spinner to the next char */
	if (++spin_cp >= (spin + SPIN_SIZE)) spin_cp = spin;

	return ch;
}

/*
 * Initialize heights and widths based on various strings and environment
 * variables (such as ENV_USE_COLOR).
 */
void
dprompt_init(struct file_node *file_list)
{
	uint8_t nls = 0;
	int max_cols, nthfile = 0, nlines = 0;
	struct file_node *curfile;

	/* Calculate width of [X]dialog(1) --gauge box */
	dwidth = label_size + pbar_size + 9;

	/* Calculate height of [X]dialog(1) --gauge box */
	dheight = 5;
	dheight += dialog_prompt_numlines(pprompt, 0) -
	    dialog_prompt_nlstate(pprompt);
	if (use_xdialog && (pprompt == NULL || *pprompt == '\0')) dheight++;
	fheight = nfiles > FDPV_DISPLAY_LIMIT ? FDPV_DISPLAY_LIMIT : nfiles;
	if (use_dialog) nls = dialog_prompt_nlstate(pprompt);
	for (curfile = file_list; curfile != NULL; curfile = curfile->next) {
		nthfile++;
		nlines += dialog_prompt_numlines(curfile->name, nls);
		if ((nthfile % FDPV_DISPLAY_LIMIT) == 0) {
			if (nlines > fheight) fheight = nlines;
			nlines = nthfile = 0;
		}
	}
	if (nlines > fheight) fheight = nlines;
	dheight += fheight;
	dheight += dialog_prompt_numlines(aprompt, 1) -
	    dialog_prompt_nlstate(aprompt);

	/* If using Xdialog(1), adjust accordingly (based on testing) */
	if (use_xdialog) dheight += dheight / 4;

	/* If wide, long prefix (`-p') or append (`-a') strings bump width */
	if (wide)
	{
		int len;

		len = (int)dialog_prompt_longestline(pprompt, 0); /* !nls */
		if ((len + 4) > dwidth) dwidth = len + 4;
		len = (int)dialog_prompt_longestline(aprompt, 1); /* nls */
		if ((len + 4) > dwidth) dwidth = len + 4;
	}

	/* Enforce width constraints to maximum values */
	max_cols = dialog_maxcols();
	if (max_cols > 0 && dwidth > max_cols) dwidth = max_cols;
	if (pbar_size > dwidth - 9) pbar_size = dwidth - 9;
	if (label_size > (dwidth - pbar_size - 9))
		label_size = dwidth - pbar_size - 9;
		/* -9 = "| " ... " - [" ... "] |" */

	/* Touch up the height (now that we know dwidth) */
	dheight += dialog_prompt_wrappedlines(pprompt, dwidth - 4, 0);
	dheight += dialog_prompt_wrappedlines(aprompt, dwidth - 4, 1);

	/* Calculate left/right portions of % and Done */
	pct_lsize = (pbar_size - 4) / 2; /* -4 == printf("%-3s%%", pct) */
	pct_rsize = pct_lsize;
	done_lsize = (pbar_size - 4) / 2; /* -4 == "Done" */
	done_rsize = done_lsize;
	/* If not evenly divisible by 2, increment the right-side */
	if ((pct_rsize + pct_rsize + 4) != pbar_size) pct_rsize++;
	if ((done_rsize + done_lsize + 4) != pbar_size) done_rsize++;

	/* Initialize "Pending" text */
	if (pbar_size < PENDING_SIZE) {
		pend_lsize = pbar_size;
		pend_rsize = 0;
		*(pending + pbar_size) = '\0';
	} else {
		*pending = '\0';
		strcat(pending, PENDING);
		/* Calculate left/right portions of mini-progressbar */
		pend_lsize = (pbar_size - PENDING_SIZE) / 2;
		pend_rsize = pend_lsize;
		/* If not evenly divisible by 2, increment the right-side */
		if ((pend_rsize + pend_lsize + 7) != pbar_size) pend_rsize++;
	}

	dprompt_clear();
}

/*
 * Clear the [X]dialog(1) `--gauge' prompt buffer.
 */
void
dprompt_clear(void)
{

	*dprompt = '\0';
	dprompt_pos = dprompt;
}

/*
 * Append to the [X]dialog(1) `--gauge' prompt buffer. Syntax is like printf(3)
 * and returns the number of bytes appended to the buffer.
 */
int
dprompt_add(const char *format, ...)
{
	int len;
	va_list ap;

	if (dprompt_pos >= (dprompt + PROMPT_MAX)) return 0;

	va_start(ap, format);
	len = vsnprintf(dprompt_pos, (size_t)(PROMPT_MAX -
	    (dprompt_pos - dprompt)), format, ap);
	va_end(ap);
	if (len == -1)
		errx(EXIT_FAILURE, "%s: Oops, dprompt buffer overflow",
		    __func__);

	if ((dprompt_pos + len) < (dprompt + PROMPT_MAX))
		dprompt_pos += len;
	else
		dprompt_pos = dprompt + PROMPT_MAX;

	return len;
}

/*
 * Process the file_node linked-list of named file-descriptors, re-generating
 * the [X]dialog(1) `--gauge' prompt text for the current state of transfers.
 */
void
dprompt_recreate(struct file_node *file_list, struct file_node *curfile,
    int pct)
{
	uint8_t after_curfile = ( curfile ? 0 : 1 );
	uint8_t nls = 0;
	int nlines = 0, nthfile = 0;
	char bold_code = 'b'; /* default: enabled */
	char color_code = '4'; /* default: blue */
	char flabel[FLABEL_MAX + 1];
	char pbar[pbar_size + 16]; /* +15 for optional color */
	struct config *option;
	struct file_node *fp;

	/* Override color defaults with that of main progress bar */
	if (use_colors || use_shadow) { /* NB: shadow enables color */
		option = dialogrc_config_option("gauge_color");
		if (option != NULL && option->value.str != NULL) {
			color_code = option->value.str[0];
			bold_code = option->value.str[2];
		}
	}

	/*
	 * Create the mini-progressbar
	 */
	snprintf(pbar, pbar_size + 1, "%*s%3u%%%*s", pct_lsize, "", pct,
	    pct_rsize, "");
	if (use_color)
	{
		int pwidth;
		char pbar_cap[sizeof(pbar)];
		char pbar_fill[sizeof(pbar)];

		/* Calculate the fill-width of progressbar */
		pwidth = pct * pbar_size / 100;
		/* Round up based on one-tenth of a percent */
		if ((pct * pbar_size % 100) > 50) pwidth++;

		/*
		 * Make two copies of pbar. Make one represent the fill and the
		 * other the remainder (cap). We'll insert the ANSI delimiter
		 * in between.
		 */
		*pbar_fill = '\0';
		*pbar_cap = '\0';
		strncat(pbar_fill, (const char *)(pbar), dwidth);
		*(pbar_fill + pwidth) = '\0';
		strncat(pbar_cap, (const char *)(pbar+pwidth), dwidth);

		/* Finalize the mini [color] progressbar */
		snprintf(pbar, sizeof(pbar), "\\Z%c\\Zr\\Z%c%s%s%s\\Zn",
		    bold_code, color_code, pbar_fill, "\\ZR", pbar_cap);
	}

	/*
	 * Build the prompt text
	 */
	dprompt_clear();
	for (fp = file_list; fp != NULL; fp = fp->next)
	{
		int flabel_size = label_size;
		char *lastline, *name = fp->name;

		nthfile++;

		/* Are we before the current file still? */
		if (fp == curfile) after_curfile = 1;

		/*
		 * Support multiline filenames (where the filename is taken as
		 * the last line and the text leading up to the last line can
		 * be used as (for example) a heading/separator between files.
		 */
		if (use_dialog) nls = dialog_prompt_nlstate(pprompt);
		nlines += dialog_prompt_numlines(name, nls);
		lastline = dialog_prompt_lastline(name, 1);
		if (name != lastline)
		{
			char c = *lastline;

			*lastline = '\0';
			dprompt_add("%s", name);
			*lastline = c;
			name = lastline;
		}

		/* Support color codes (for dialog(1)) in file names */
		if (use_dialog && use_color)
		{
			char *cp = name;

			while (*cp != '\0') {
				if (*cp == '\\' && *(cp + 1) != '\0' &&
				    *(++cp) == 'Z' && *(cp + 1) != '\0')
				{
					cp++;
					flabel_size += 3;
				}
				cp++;
			}
			if (flabel_size > FLABEL_MAX)
				flabel_size = FLABEL_MAX;
		}

		/* If name is too long, add an ellipsis */
		if (snprintf(flabel, flabel_size + 1, "%s", name) >
		    flabel_size) sprintf(flabel + flabel_size - 3, "...");

		/*
		 * Append the label (processing the current file differently)
		 */
		if (fp == curfile && pct < 100)
		{
			char *cp;

			/*
			 * Add an ellipsis to current file name if it will fit.
			 * There may be an ellipsis already from truncating the
			 * label (in which case, we already have one).
			 */
			cp = flabel + strlen(flabel);
			if (cp < (flabel + flabel_size))
				snprintf(cp, flabel_size -
				    (cp - flabel) + 1, "...");

			/* Append label (with spinner and optional color) */
			dprompt_add("%s%-*s%s %c", (use_color ? "\\Zb" : ""),
			    flabel_size, flabel, (use_color ? "\\Zn" : ""),
			    spin_char());
		} else
			dprompt_add("%-*s%s %s", flabel_size,
			    flabel, (use_color ? "\\Zn" : "" ), " ");

		/*
		 * Append pbar/status (processing the current file differently)
		 */
		if (pbar_size >= 0)
		{
			if (fp == curfile && pct < 100)
				dprompt_add(" [%s]\\n", pbar);
			else if (fp != curfile && after_curfile)
				dprompt_add(" [%-*s%s%-*s]\\n", pend_lsize, "",
				    pending, pend_rsize, "");
			else if (pbar_size >= 4) {
				if (use_color)
					dprompt_add(
					    " [\\Z%c\\Zr\\Z2%-*s%s%-*s\\Zn]\\n",
					    bold_code, done_lsize, "", "Done",
					    done_rsize, "");
				else
					dprompt_add(" [%-*s%s%-*s]\\n",
					    done_lsize, "", "Done",
					    done_rsize, "");
			} else if (pbar_size < 4)
				dprompt_add(" [%s%-*s%s]\\n",
				    (use_color ? "\\Z2\\Zr" : ""),
				    pbar_size, "",
				    (use_color ? "\\Zn" : ""));
		} else
			dprompt_add("\\n");

		/* Stop building if we've hit the internal limit */
		if (nthfile >= FDPV_DISPLAY_LIMIT) break;
	}

	/*
	 * Since we cannot change the height/width of the [X]dialog(1) widget
	 * after spawn, to make things look nice let's pad the height so that
	 * the `-a text' always appears in the same spot.
	 *
	 * NOTE: fheight is calculated in dprompt_init(). It represents the
	 * maximum height required to display the set of items (broken up into
	 * pieces of FDPV_DISPLAY_LIMIT chunks) whose names contain the most
	 * newlines for any given set.
	 */
	while (nlines <= fheight)
	{
		dprompt_add("\n");
		nlines++;
	}

	/* Xdialog(1) requires newlines (a) escaped and (b) in triplicate */
	if (use_xdialog)
	{
		size_t len;

		/* Replace `\n' with `\n\\n\n' in dprompt */
		len = strlen(dprompt);
		len += strcount(dprompt, "\\n") * 5; /* +5 chars per count */
		if (len > PROMPT_MAX)
			errx(EXIT_FAILURE, "%s: Oops, dprompt buffer overflow "
			    "(%i > %i)", __func__, len, PROMPT_MAX);
		if (replaceall(dprompt, "\\n", "\n\\n\n") < 0)
			err(EXIT_FAILURE, "%s: replaceall()", __func__);
	}
}

/*
 * Print the [X]dialog(1) `--gauge' prompt text to a buffer.
 */
int
dprompt_sprint(char * restrict str, const char *prefix, const char *append)
{

	return snprintf(str, PROMPT_MAX, "%s%s%s%s", use_color ? "\\Zn" : "",
	    prefix ? prefix : "", dprompt, append ? append : "");
}

/*
 * Print the [X]dialog(1) `--gauge' prompt text to file descriptor fd (could
 * be STDOUT_FILENO or a pipe(2) file descriptor to actual [X]dialog(1)).
 */
void
dprompt_dprint(int fd, const char *prefix, const char *append, int overall)
{

	dprintf(fd, "XXX\n%s%s%s%s\nXXX\n%i\n", use_color ? "\\Zn" : "",
	    prefix ? prefix : "", dprompt, append ? append : "", overall);
	fsync(fd);
}
