/*-
 * Copyright (c) 2013 Devin Teske <dteske@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <limits.h>
#define _BSD_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include "fdpv.h"
#include "dprompt.h"
#include "string_m.h"
#include "dialog.h"

/* Test Mechanics (used with `-T') */
#define INCREMENT		1	/* How much to increment % each time */
#define XDIALOG_INCREMENT	15	/* different for slower Xdialog(1) */
#define SLEEP_USEC		0	/* ms to sleep between increments */
#define XDIALOG_SLEEP_USEC	0	/* different for slower Xdialog(1) */
static uint8_t increment = INCREMENT;
static useconds_t sleep_usec = SLEEP_USEC;

/* Debugging */
uint8_t debug = 0;

/* Data to process */
static int dtablesize;
unsigned int nfiles = 0;
static struct file_node *file_list = NULL;

/* Extra display information */
uint8_t wide = 0;	/* `-w' */
char *pgm;		/* set to argv[0] by main() */
char *aprompt = NULL;	/* `-a text' */
char *pprompt = NULL;	/* `-p text' */

/* Function prototypes for private functions (see style(9)) */
void usage(void);

/*
 * Takes a list of names that are to correspond to input streams coming from
 * numbered file descriptors (starting at 3 for the first name) and produces
 * necessary output to drive a[n] [X]dialog(1) `--gauge' widget. If the `-d'
 * flag is used, output is instead send to terminal standard output (and the
 * output can then be saved to a file, piped into custom [X]dialog(1) invoc-
 * ation, or other.
 */
int
main(int argc, char *argv[])
{
	int ch, out = STDOUT_FILENO;
	int nthfile = 0, overall = 0;
	pid_t pid;
	size_t len;
	char *cp;
	struct file_node *list_head, *curfile;

	pgm = argv[0]; /* store a copy of invocation name */

	/* Get maximum number of file descriptors */
	dtablesize = getdtablesize();

	/*
	 * Process command-line options
	 */
	while ((ch = getopt(argc, argv, "a:b:dhL:p:P:t:TwX")) != -1)
	{
		switch(ch) {
		case 'a': /* additional message text to append */
			if (aprompt == NULL)
				if ((aprompt = malloc(APROMPT_MAX)) == NULL)
					errx(EXIT_FAILURE, "Out of memory?!");
			snprintf(aprompt, APROMPT_MAX, "%s", optarg);
			break;
		case 'b': /* [X]dialog(1) backtitle */
			if (backtitle != NULL) free(backtitle);
			if ((backtitle = malloc(strlen(optarg) + 1)) == NULL)
				errx(EXIT_FAILURE, "Out of memory?!");
			*backtitle = '\0';
			strcat(backtitle, optarg);
			break;
		case 'd': /* Debugging */
			debug = 1;
			break;
		case 'h': /* Help/Usage */
			usage();
			break;
		case 'L': /* custom label size */
			label_size = (int)strtol(optarg, (char **)NULL, 10);
			if (label_size == 0 && errno == EINVAL)
				errx(EXIT_FAILURE,
				    "`-L' argument must be numeric");
			else if (label_size < -1)
				label_size = -1;
			break;
		case 'p': /* additional message text to use as prefix */
			if (pprompt == NULL)
				if ((pprompt = malloc(PPROMPT_MAX +2)) == NULL)
					errx(EXIT_FAILURE, "Out of memory?!");
				/* +2 is for implicit "\n" appended later */
			snprintf(pprompt, PPROMPT_MAX, "%s", optarg);
			break;
		case 'P': /* custom size for mini-progressbar */
			pbar_size = (int)strtol(optarg, (char **)NULL, 10);
			if (pbar_size == 0 && errno == EINVAL)
				errx(EXIT_FAILURE,
				    "`-P' argument must be numeric");
			else if (pbar_size < -1)
				pbar_size = -1;
			break;
		case 't': /* [X]dialog(1) title */
			if (title != NULL) free(title);
			if ((title = malloc(strlen(optarg) + 1)) == NULL)
				errx(EXIT_FAILURE, "Out of memory?!");
			*title = '\0';
			strcat(title, optarg);
			break;
		case 'T': /* test mode (don't read data, fake it) */
			dialog_test = 1;
			break;
		case 'w': /* `-p' and `-a' widths bump [X]dialog(1) width */
			wide = 1;
			break;
		case 'X': /* X11 support through x11/xdialog */
			snprintf(dialog, PATH_MAX, XDIALOG);
			use_xdialog = 1;
			use_dialog = 0;
			use_color = 0;
			break;
		case '?': /* unknown argument (based on optstring) */
		default: /* unhandled argument (based on switch) */
			usage();
		}
	}
	argc -= optind;
	argv += optind;

	/* Process the contents of DIALOGRC (~/.dialogrc) */
	if (use_dialog) {
		int res;
		res = parse_dialogrc();
		if (debug && res == 0) {
			warnx("Successfully read `%s' config file", DIALOGRC);
			warnx("use_shadow = %i (Boolean)", use_shadow);
			warnx("use_colors = %i (Boolean)", use_colors);
			warnx("gauge_color=[%s] (FGH)", gauge_color);
		}
	}

	/* Process remaining arguments as list of names to display */
	{
		uint8_t pprompt_nls = 0; /* See dialog_prompt_nlstate() */
		uint8_t shrink_label_size = 0;
		int max_cols, n = 0;
		size_t size = sizeof(struct file_node);

		if (use_dialog && pprompt != NULL && *pprompt != '\0')
			pprompt_nls = dialog_prompt_nlstate(pprompt);

		max_cols = dialog_maxcols();

		if (label_size == -1) shrink_label_size = 1;
		for (curfile = file_list; n < argc; n++)
		{
			nfiles++;

			/* Allocate a new struct for the file argument */
			if (curfile == NULL) {
				if ((curfile = malloc(size)) == NULL)
					errx(EXIT_FAILURE, "Out of memory?!");
				memset((void *)(curfile), '\0', size);
				file_list = curfile;
			} else {
				if ((curfile->next = malloc(size)) == NULL)
					errx(EXIT_FAILURE, "Out of memory?!");
				memset((void *)(curfile->next), '\0', size);
				curfile = curfile->next;
			}
			curfile->name = argv[n];
			curfile->fd = (int)(nfiles + 2);

			/* Optionally calculate label size for file */
			if (shrink_label_size)
			{
				int nls = 0; /* See dialog_prompt_nlstate() */
				char *last, *name = argv[n];

				if (curfile == file_list) nls = pprompt_nls;
				last = dialog_prompt_lastline(name, nls);
				if (use_dialog) {
					char c = *last;

					*last= '\0';
					nls = dialog_prompt_nlstate(name);
					*last = c;
				}
				len = dialog_prompt_longestline(last, nls);
				if ((int)len > (label_size - 3)) {
					if (label_size > 0) label_size += 3;
					label_size = len;
					/* Room for ellipsis (unless NULL) */
					if (label_size > 0) label_size += 3;
				}

				if (max_cols > 0 && label_size >
				    (max_cols - pbar_size - 9))
					label_size = max_cols - pbar_size - 9;
			}
		}
	}

	/* If $USE_COLOR is set and non-NULL enable color; otherwise disable */
	if ((cp = getenv(ENV_USE_COLOR)) != 0)
		use_color = ( *cp != '\0' ? 1 : 0 );

	/* Display usage and exit if not given at least one name */
	if (nfiles == 0) {
		warnx("no file names provided");
		usage();
	} else if (debug)
		warnx("%u file name%s provided (as many descriptors expected)",
		    nfiles, (nfiles != 1 ? "s" : ""));

	/* Print some debugging information */
	if (debug)
	{
		warnx("getdtablesize() = %i (max number of file descriptors)",
		    dtablesize);
		warnx("%s(1) max rows x cols = %i x %i",
		    (use_xdialog ? XDIALOG : DIALOG),
		    dialog_maxrows(), dialog_maxcols());
	}

	/* Xdialog(1) updates a lot slower than dialog(1) */
	if (dialog_test && use_xdialog) increment = XDIALOG_INCREMENT;

	/* Always add implicit newline to pprompt (when specified) */
	if (pprompt != NULL && *pprompt != '\0')
	{
		len = strlen(pprompt);
		/*
		 * NOTE: pprompt = malloc(PPROMPT_MAX + 2)
		 * NOTE: (see getopt(2) section above for pprompt allocation)
		 */
		pprompt[len++] = '\\';
		pprompt[len++] = 'n';
		pprompt[len++] = '\0';
	}

	/* Xdialog(1) requires newlines (a) escaped and (b) in triplicate */
	if (use_xdialog && pprompt != NULL)
	{
		/* Replace `\n' with `\n\\n\n' in pprompt */
		len = strlen(pprompt);
		len += strcount(pprompt, "\\n") * 2;
		if (len > PPROMPT_MAX)
			errx(EXIT_FAILURE, "%s: Oops, pprompt buffer overflow "
			    "(%i > %i)", __func__, len, PPROMPT_MAX);
		if (replaceall(pprompt, "\\n", "\n\\n\n") < 0)
			err(EXIT_FAILURE, "%s: replaceall()", __func__);
	}
	if (use_xdialog && aprompt != NULL)
	{
		/* Replace `\n' with `\n\\n\n' in aprompt */
		len = strlen(aprompt);
		len += strcount(aprompt, "\\n") * 2;
		if (len > APROMPT_MAX)
			errx(EXIT_FAILURE, "%s: Oops, aprompt buffer overflow "
			    " (%i > %i)", __func__, len, APROMPT_MAX);
		if (replaceall(aprompt, "\\n", "\n\\n\n") < 0)
			err(EXIT_FAILURE, "%s: replaceall()", __func__);
	}

	/*
	 * Warn user about an obscure dialog(1) bug (Xdialog(1) not affected)
	 * in the `--gauge' widget. If the first non-whitespace letter of
	 * "{new_prompt}" in the "XXX\n{new_prompt}\nXXX\n" I/O is a number,
	 * the number can sometimes be mistaken for a percentage to the overall
	 * progressbar. Other nasty side-effects such as the entire prompt not
	 * displaying or displaying improperly are caused by this bug too.
	 *
	 * NOTE: When we can use color, we have a work-around... prefix the
	 * output with `\Zn' (used to terminate ANSI and reset to normal).
	 */
	if (use_dialog && !use_color)
	{
		int backslash = 0;
		char *fc;

		/* First, check pprompt (falls through if NULL) */
		fc = pprompt;
		while (fc != NULL && *fc != '\0')
		{
			if (*fc == '\n') /* leading literal newline OK */
				break;
			if (!isspace(*fc) && *fc != '\\' && backslash == 0)
				break;
			else if (backslash > 0 && *fc != 'n')
				break;
			else if (*fc == '\\') {
				backslash++;
				if (backslash > 2) break; /* we're safe */
			}
			fc++;
		}
		/* First non-whitespace character that dialog(1) will see */
		if (fc != NULL && *fc >= '0' && *fc <= '9')
			warnx("WARNING! text argument to `-p' begins with a "
			    "number (not recommended)");
		else if (fc > pprompt)
			warnx("WARNING! text argument to `-p' begins with "
			    "whitespace (not recommended)");

		/*
		 * If no pprompt or pprompt is all whitespace, check the first
		 * file name provided to make sure it is alright too.
		 */
		if ((pprompt == NULL || *fc == '\0') && file_list != NULL)
		{
			struct file_node *first_file = file_list;

			fc = first_file->name;
			while (fc != NULL && *fc != '\0' && isspace(*fc)) fc++;
			/* First non-whitespace char that dialog(1) will see */
			if (fc != NULL && *fc >= '0' && *fc <= '9')
				warnx("WARNING! File name `%s' begins with "
				    "a number (use `-p text' for safety)",
				    first_file->name);
		}
	}

	/* Check to make sure we haven't been given too many files */
	if (nfiles > (unsigned int)dtablesize - 3)
		errx(EXIT_FAILURE, "ERROR! Cannot handle more than %i streams",
		    dtablesize);

	dprompt_init(file_list);
		/* Reads: label_size, pbar_size, pprompt, aprompt, nfiles */
		/* Inits: dheight and dwidth */

	if (!debug)
	{
		char init_prompt[PROMPT_MAX + 1] = "";

		/* Internally create the initial `--gauge' prompt text */
		dprompt_recreate(file_list, (struct file_node *)NULL, 0);

		/* Spawn [X]dialog(1) `--gauge', returning pipe descriptor */
		dprompt_sprint(init_prompt, pprompt, aprompt);
		out = dialog_spawn_gauge(init_prompt, &pid);
		dprompt_dprint(out, pprompt, aprompt, 0);
	}

	/*
	 * Process the file streams [serially] (one for each argument passed)
	 */
	list_head = file_list;
	for (curfile = file_list; curfile != NULL; curfile = curfile->next)
	{
		int pct = 0;

		nthfile++;

		if (dialog_test)
			pct = 0 - increment;

		while (pct < 100)
		{
			if (dialog_test) {
				if (sleep_usec) usleep(sleep_usec);
				pct += increment;
			} else
				/* XXX Pretend file is fully-processed XXX */
				pct = 100;

			/* Calculate total overall progress (rounding up) */
			overall = (100 * nthfile - 100 + pct) / nfiles;
			if (((100 * nthfile - 100 + pct) * 10 / nfiles % 100) >
			    50) overall++;

			dprompt_recreate(list_head, curfile, pct);
			dprompt_dprint(out, pprompt, aprompt, overall);
			fsync(out);
		}

		/* Advance head of list when we hit the max display lines */
		if (nthfile % FDPV_DISPLAY_LIMIT == 0)
			list_head = curfile->next;
	}

	if (!debug)
	{
		close(out);
		waitpid(pid, (int *)NULL, 0);	
	}

	exit(EXIT_SUCCESS);
}

/*
 * Print short usage statement to stderr and exit with error status.
 */
void
usage(void)
{

	if (debug) exit(EXIT_FAILURE); /* No need for usage */

	fprintf(stderr, "Usage: %s [-dhTwX] [-a text] [-b backtitle]"
	    " [-L size] [-p text] [-P size ]\n", pgm);
	fprintf(stderr, "       %*s [-t title] file ...\n", strlen(pgm), "");
	fprintf(stderr, "OPTIONS:\n");
#define OPTFMT "\t%-15s %s\n"
	fprintf(stderr, OPTFMT, "-a text",
	    "Append text. Displayed below file progress indicators.");
	fprintf(stderr, OPTFMT, "-b backtitle",
	    "String to be displayed on the backdrop, at top-left.");
	fprintf(stderr, OPTFMT, "-d",
	    "Debug. Write to standard output instead of dialog(1).");
	fprintf(stderr, OPTFMT, "-h",
	    "Produce this output on standard error and exit.");
	fprintf(stderr, OPTFMT, "-L size",
	    "Label size. Must be a number greater than 0, or -1.");
	fprintf(stderr, OPTFMT, "-p text",
	    "Prefix text. Displayed above file progress indicators.");
	fprintf(stderr, OPTFMT, "-P size",
	    "Mini-progressbar size. Must be a number greater than 3.");
	fprintf(stderr, OPTFMT, "-t title",
	    "Title string to be displayed at top of dialog(1) box.");
	fprintf(stderr, OPTFMT, "-T",
	    "Test mode. Don't actually read any data, but fake it.");
	fprintf(stderr, OPTFMT, "-w",
	    "Wide. Width of `-p' and `-a' text bump dialog(1) width.");
	fprintf(stderr, OPTFMT, "-X",
	    "X11. Use Xdialog(1) instead of dialog(1).");
	exit(EXIT_FAILURE);
}
